import inspect
import typing

from robotpy_installer import pyproject, pypackages
from robotpy_installer.installer import _WPILIB_YEAR as YEAR

from packaging.requirements import Requirement


def load_project(content: str) -> pyproject.RobotPyProjectToml:
    return pyproject.loads(inspect.cleandoc(content))


def null_resolver(req: Requirement, env: pypackages.Env) -> typing.List[Requirement]:
    return []


def test_ok():
    project = load_project(
        f"""
        [tool.robotpy]
        robotpy_version = "{YEAR}.1.1.2"
    """
    )
    installed = pypackages.make_packages({"robotpy": f"{YEAR}.1.1.2"})
    assert project.are_requirements_met(
        installed, pypackages.roborio_env(), null_resolver
    ) == (
        True,
        [],
    )


def test_older_fail():
    project = load_project(
        f"""
        [tool.robotpy]
        robotpy_version = "{YEAR}.1.1.2"
    """
    )
    installed = pypackages.make_packages({"robotpy": f"{YEAR}.1.1.0"})
    assert project.are_requirements_met(
        installed, pypackages.roborio_env(), null_resolver
    ) == (
        False,
        [f"robotpy=={YEAR}.1.1.2 (found {YEAR}.1.1.0)"],
    )


def test_older_and_newer_fail():
    project = load_project(
        f"""
        [tool.robotpy]
        robotpy_version = "{YEAR}.1.1.2"
    """
    )
    installed = pypackages.make_packages(
        {"robotpy": [f"{YEAR}.1.1.0", f"{YEAR}.1.1.4"]}
    )
    assert project.are_requirements_met(
        installed, pypackages.roborio_env(), null_resolver
    ) == (
        False,
        [f"robotpy=={YEAR}.1.1.2 (found {YEAR}.1.1.0, {YEAR}.1.1.4)"],
    )


def test_beta_empty_req():
    project = load_project(
        f"""
        [tool.robotpy]
        robotpy_version = "{YEAR}.1.1.2"
        requires = [
            "robotpy-commands-v2"
        ]
    """
    )

    installed = pypackages.make_packages(
        {"robotpy": f"{YEAR}.1.1.2", "robotpy-commands-v2": f"{YEAR}.0.0b4"}
    )

    assert project.are_requirements_met(
        installed, pypackages.roborio_env(), null_resolver
    ) == (
        True,
        [],
    )


def test_env_marker():
    project = load_project(
        f"""
        [tool.robotpy]
        robotpy_version = "{YEAR}.1.1.2"
        requires = [
            "robotpy-opencv; platform_machine == 'roborio'",
            "opencv-python; platform_machine != 'roborio'"
        ]
    """
    )

    installed = pypackages.make_packages(
        {"robotpy": f"{YEAR}.1.1.2", "robotpy-opencv": f"{YEAR}.0.0"}
    )

    assert project.are_requirements_met(
        installed, pypackages.roborio_env(), null_resolver
    ) == (
        True,
        [],
    )



import dataclasses
from importlib.metadata import metadata, PackageNotFoundError
import inspect
import pathlib
import typing

from packaging.requirements import Requirement
from packaging.version import Version, InvalidVersion
import tomli
import tomlkit

from . import installer
from . import pypackages
from .pypackages import Packages, Env
from .errors import Error


class PyprojectError(Error):
    pass


class NoRobotpyError(PyprojectError):
    pass


class UnsupportedRobotpyVersion(PyprojectError):
    pass


def toml_path(project_path: pathlib.Path):
    return project_path / "pyproject.toml"


@dataclasses.dataclass
class RobotPyProjectToml:
    """
    The results of parsing a ``pyproject.toml`` file in a RobotPy project. This
    only parses fields that matter for deploy/install (TODO: this probably
    should live in a single unified place?)

    .. code-block:: toml

        [tool.robotpy]

        # equivalent to `robotpy==2024.0.0b4`
        robotpy_version = "2024.0.0b4"

        # equivalent to `robotpy[cscore, ...]`
        robotpy_extras = ["cscore"]

        # Other pip installable requirement lines
        requires = [
        "numpy"
        ]

    """

    #: Version of robotpy that is depended on
    robotpy_version: Version

    robotpy_extras: typing.List[str] = dataclasses.field(default_factory=list)

    #: Requirement for the robotpy meta package -- all RobotPy projects must
    #: depend on it
    @property
    def robotpy_requires(self) -> Requirement:
        if self.robotpy_extras:
            extras = f"[{','.join(self.robotpy_extras)}]"
        else:
            extras = ""
        return Requirement(f"robotpy{extras}=={self.robotpy_version}")

    #: User's custom requirements
    requires: typing.List[Requirement] = dataclasses.field(default_factory=list)

    def are_requirements_met(
        self,
        packages: Packages,
        env: Env,
        extra_resolver: pypackages.ExtraResolver,
    ) -> typing.Tuple[bool, typing.List[str]]:
        """
        Determines if the set of packages meets the requirements specified by
        this project
        """
        reqs = self.get_install_reqs()
        assert reqs and reqs[0].name == "robotpy"
        robotpy_req = reqs[0]

        # Extra requirements from the extra resolver
        reqs.extend(extra_resolver(robotpy_req, env))

        return pypackages.are_requirements_met(reqs, packages, env)

    def are_local_requirements_met(
        self,
    ) -> typing.Tuple[bool, typing.List[str]]:
        """
        Determines if the locally installed packages meets the requirements
        specified by this project
        """

        return self.are_requirements_met(
            pypackages.get_local_packages(), {}, pypackages.extra_resolver_local
        )

    def get_install_reqs(self) -> typing.List[Requirement]:
        return [self.robotpy_requires] + self.requires

    def get_install_list(self) -> typing.List[str]:
        return list(map(str, self.get_install_reqs()))


def robotpy_installed_version() -> str:
    # this is a bit weird because this project doesn't depend on robotpy, it's
    # the other way around.. but oh well?
    try:
        return metadata("robotpy")["Version"]
    except PackageNotFoundError:
        raise NoRobotpyError(
            "cannot infer default robotpy package version: robotpy package not installed "
            "(do `pip install robotpy` or create a pyproject.toml)"
        ) from None


def write_default_pyproject(
    project_path: pathlib.Path,
):
    """
    Using the current environment, write a minimal pyproject.toml

    :param project_path: Path to robot project
    """

    robotpy_version = robotpy_installed_version()

    provides_extra = metadata("robotpy").get_all("Provides-Extra")
    if not provides_extra:
        extras = ""
    else:
        extras = "\n    # ".join(f'"{extra}",' for extra in sorted(provides_extra))

    content = inspect.cleandoc(
        f"""
            
            #
            # Use this configuration file to control what RobotPy packages are installed
            # on your RoboRIO
            #

            [tool.robotpy]

            # Version of robotpy this project depends on
            robotpy_version = "{robotpy_version}"
            
            # Which extra RobotPy components should be installed
            # -> equivalent to `pip install robotpy[extra1, ...]
            robotpy_extras = [
                # ##EXTRAS##
            ]

            # Other pip packages to install
            requires = []

        """
    )

    content += "\n"
    content = content.replace("##EXTRAS##", extras)

    with open(toml_path(project_path), "w") as fp:
        fp.write(content)


def load(
    project_path: pathlib.Path,
    *,
    write_if_missing: bool = False,
    default_if_missing=False,
) -> RobotPyProjectToml:
    """
    Reads a pyproject.toml file for a RobotPy project. Raises FileNotFoundError
    if the file isn't present

    :param project_path: Path to robot project
    """

    pyproject_path = toml_path(project_path)
    if not pyproject_path.exists():
        if default_if_missing:
            return RobotPyProjectToml(
                robotpy_version=Version(robotpy_installed_version())
            )
        if write_if_missing:
            write_default_pyproject(project_path)

    with open(pyproject_path, "rb") as fp:
        data = tomli.load(fp)

    return _load(str(pyproject_path), data)


def loads(content: str):
    data = tomli.loads(content)
    return _load("<string>", data)


def _load(
    pyproject_path: str, data: typing.Dict[str, typing.Any]
) -> RobotPyProjectToml:
    try:
        robotpy_data = data["tool"]["robotpy"]
        if not isinstance(robotpy_data, dict):
            raise KeyError()
    except KeyError:
        raise PyprojectError(
            f"{pyproject_path} must have [tool.robotpy] section"
        ) from None

    try:
        robotpy_version = Version(robotpy_data["robotpy_version"])
    except KeyError:
        raise PyprojectError(
            f"{pyproject_path} missing required tools.robotpy.robotpy_version"
        ) from None
    except InvalidVersion:
        raise PyprojectError(
            f"{pyproject_path}: tools.robotpy.robotpy_version is not a valid version"
        ) from None

    supported_year = int(installer._WPILIB_YEAR)
    if robotpy_version.major != supported_year:
        msg = (
            f"Only RobotPy {supported_year}.x is supported by this version "
            f"of robotpy-installer ({pyproject_path} has {robotpy_version})"
        )
        raise UnsupportedRobotpyVersion(msg)

    robotpy_extras_any = robotpy_data.get("robotpy_extras")
    if isinstance(robotpy_extras_any, list):
        robotpy_extras = list(map(str, robotpy_extras_any))
    elif not robotpy_extras_any:
        robotpy_extras = []
    else:
        robotpy_extras = [str(robotpy_extras_any)]

    requires_any = robotpy_data.get("requires")
    if isinstance(requires_any, list):
        requires = []
        for req in requires_any:
            requires.append(Requirement(req))
    elif requires_any:
        requires = [Requirement(str(requires_any))]
    else:
        requires = []

    return RobotPyProjectToml(
        robotpy_version=robotpy_version,
        robotpy_extras=robotpy_extras,
        requires=requires,
    )


def set_robotpy_version(project_path: pathlib.Path, version: Version):
    pyproject_path = toml_path(project_path)
    with open(pyproject_path) as fp:
        data = tomlkit.parse(fp.read())

    try:
        data["tool"]["robotpy"]["robotpy_version"] = str(version)  # type: ignore
    except Exception as e:
        raise ValueError("`pyproject.toml` is not valid") from e

    rawdata = tomlkit.dumps(data)

    with open(pyproject_path, "w") as fp:
        fp.write(rawdata)



"""
Various python packaging related logic
"""

from importlib.metadata import distributions, metadata, PackageNotFoundError
import pathlib
import typing
import zipfile

from packaging.metadata import Metadata
from packaging.requirements import Requirement
from packaging.utils import (
    canonicalize_name,
    parse_sdist_filename,
    parse_wheel_filename,
    NormalizedName,
    InvalidSdistFilename,
    InvalidWheelFilename,
)
from packaging.version import Version

#: environment markers needed by Marker.evaluate
Env = typing.Dict[str, str]

Packages = typing.Dict[NormalizedName, typing.List[Version]]

ExtraResolver = typing.Callable[[Requirement, Env], typing.List[Requirement]]


def are_requirements_met(
    requirements: typing.List[Requirement],
    packages: Packages,
    env: Env,
) -> typing.Tuple[bool, typing.List[str]]:
    """
    Given a set of packages and a list of requirements, determine if the packages
    satisfy the list of requirements
    """

    unmet_requirements = []

    for req in requirements:
        # Ignore this requirement if it doesn't apply to the specified
        # environment
        if req.marker and not req.marker.evaluate(env):
            continue

        req_name = canonicalize_name(req.name)
        req.specifier.prereleases = True

        empty_specifier = str(req.specifier) == ""

        for pkg, pkg_versions in packages.items():
            if pkg == req_name:
                if not empty_specifier:
                    for pkg_version in pkg_versions:
                        if pkg_version in req.specifier:
                            break
                    else:
                        found = ", ".join(map(str, sorted(pkg_versions)))
                        unmet_requirements.append(
                            f"{req.name}{req.specifier} (found {found})"
                        )
                break
        else:
            unmet_requirements.append(f"{req.name}{req.specifier} (not found)")

    return not bool(unmet_requirements), unmet_requirements


def evaluate_extras_markers(
    reqs: typing.List[Requirement], env: Env, extras: typing.Iterable[str]
) -> typing.List[Requirement]:
    env = env.copy()
    matched = []
    for req in reqs:
        for extra in extras:
            env["extra"] = extra
            if req.marker is None or req.marker.evaluate(env):
                # marker is no longer needed
                req.marker = None
                matched.append(req)
                break
    return matched


def extra_resolver_local(req: Requirement, env: Env) -> typing.List[Requirement]:
    """
    Given a requirement, resolves its extras using the version of the
    requirement installed locally. Ignores its markers.

    Fails silently.
    """
    if not req.extras:
        return []

    try:
        m = metadata(req.name)
    except PackageNotFoundError:
        return []

    extra_reqs = []
    requires_dist = m.get_all("Requires-Dist")
    if requires_dist:
        extra_reqs = evaluate_extras_markers(
            [Requirement(r) for r in requires_dist], env, req.extras
        )

    return extra_reqs


def make_cache_extra_resolver(packages: Packages) -> ExtraResolver:
    """
    :param packages: The list of packages in the cache as returned by
                     get_pip_cache_packages
    """

    def _resolver(req: Requirement, env: Env) -> typing.List[Requirement]:
        if not req.extras:
            return []

        env = env.copy()
        env["extra"] = ",".join(req.extras)

        # Find the requirement
        creqs = packages.get(canonicalize_name(req.name))
        if creqs is None:
            raise KeyError(f"{req} not downloaded in cache (did you do a sync?)")

        req.specifier.prereleases = True
        for creq in sorted(creqs, reverse=True):
            if req.specifier is None or creq in req.specifier:
                break
        else:
            raise KeyError(f"{req} not downloaded in cache (did you do a sync?)")

        if not isinstance(creq, CacheVersion):
            raise ValueError("internal error")

        m = metadata_from_wheel(creq.file_path)
        return evaluate_extras_markers(m.requires_dist, env, req.extras)

    return _resolver


def get_local_packages() -> Packages:
    """
    Iterates over locally installed packages and returns dict of versions
    """
    return {
        canonicalize_name(dist.metadata["Name"]): [Version(dist.version)]
        for dist in distributions()
    }


class CacheVersion(Version):
    def __init__(self, version: str, file_path: pathlib.Path) -> None:
        super().__init__(version)
        self.file_path = file_path


def get_pip_cache_packages(
    cache_root: pathlib.Path,
) -> Packages:
    """
    Iterates over the pip cache and returns dict of packages
    """

    packages: Packages = {}

    for f in (cache_root / "pip_cache").iterdir():
        if f.suffix == ".whl":
            try:
                name, version, _, _ = parse_wheel_filename(f.name)
                packages.setdefault(name, []).append(CacheVersion(str(version), f))
            except InvalidWheelFilename:
                pass
        elif f.suffix in (".gz", ".zip"):
            try:
                name, version = parse_sdist_filename(f.name)
                packages.setdefault(name, []).append(CacheVersion(str(version), f))
            except InvalidSdistFilename:
                pass

    return packages


def make_packages(
    packages: typing.Mapping[str, typing.Union[typing.List[str], str]]
) -> Packages:
    """
    For unit testing
    """
    return {
        canonicalize_name(name): (
            [Version(version)]
            if isinstance(version, str)
            else [Version(v) for v in version]
        )
        for name, version in packages.items()
    }


def metadata_from_wheel(whl_path: pathlib.Path) -> Metadata:
    """
    Retrieves the metadata from a wheel file
    """
    name, version, _, _ = parse_wheel_filename(whl_path.name)
    with zipfile.ZipFile(whl_path) as zfp:
        m = zfp.read(f"{name}-{version}.dist-info/METADATA")

    return Metadata.from_email(m, validate=False)


def roborio_env() -> Env:
    """
    For use with ``packaging.marker.Marker.evaluate``
    """
    return {
        "implementation_name": "cpython",
        "implementation_version": "3.12.1",
        "os_name": "posix",
        "platform_machine": "roborio",
        "platform_python_implementation": "CPython",
        "platform_system": "Linux",
        "python_full_version": "3.12.0",
        "python_version": "3.12",
        "sys_platform": "linux",
    }



import argparse
import inspect
import logging
import pathlib

from . import pyproject
from .utils import handle_cli_error


logger = logging.getLogger("init")


class Init:
    """
    Initializes a robot project
    """

    def __init__(self, parser: argparse.ArgumentParser):
        pass

    @handle_cli_error
    def run(self, main_file: pathlib.Path, project_path: pathlib.Path):
        project_path.mkdir(parents=True, exist_ok=True)

        # Create robot.py if it doesn't already exist
        # - TODO: it would be neat if it could download an example from github
        if not main_file.exists():
            with open(main_file, "w") as fp:
                fp.write("# TODO: insert robot code here\n")

            logger.info("Created empty %s", main_file)

        # Create pyproject.toml if it doesn't already exist
        pyproject_path = pyproject.toml_path(project_path)
        if not pyproject_path.exists():
            pyproject.write_default_pyproject(project_path)

            logger.info("Created %s", pyproject_path)



import hashlib
import io
import json
import logging
import pathlib
import typing

from .sshcontroller import SshController

logger = logging.getLogger("robotpy.installer")


java_jars = "/home/lvuser/*.jar"
cpp_exe = "/home/lvuser/frcUserProgram"
robot_command = "/home/lvuser/robotCommand"
static_deploy = "/home/lvuser/deploy"

third_party_libs = "/usr/local/frc/third-party/lib"

kill_robot_script = "/usr/local/frc/bin/frcKillRobot.sh"
kill_robot_cmd = f"{kill_robot_script} -t"
kill_script_content: typing.Optional[bytes] = None


def uninstall_cpp_java_lvuser(ssh: SshController) -> bool:
    """
    Frees up disk space by removing FRC C++/Java programs. This runs as lvuser or admin.

    :returns: True if success, False if uninstall_cpp_java_admin needs to be ran
    """

    logger.info("Clearing FRC C++/Java user programs if present")

    rm_paths = (java_jars, cpp_exe, robot_command)

    ssh.exec_bash(
        ". /etc/profile.d/frc-path.sh",
        ". /etc/profile.d/natinst-path.sh",
        "set -x",
        # Kill code only if java jar present
        f"[ ! -f {java_jars} ] || {kill_robot_cmd}",
        # Kill code only if cpp exe present
        f"[ ! -f {cpp_exe} ] || {kill_robot_cmd}",
        f"rm -rf {' '.join(rm_paths)}",
    )

    # Check if admin pieces need to run
    result = ssh.exec_bash(
        '[ -z "$(opkg list-installed frc*-openjdk-*)" ]',
        f'[ ! -d {third_party_libs} ] || [ -z "$(ls /usr/local/frc/third-party/lib)" ]',
        # This is copied with admin privs, can't delete as lvuser
        f"[ ! -d {static_deploy} ]",
    )
    return result.returncode == 0


def uninstall_cpp_java_admin(ssh: SshController):
    """
    Frees up disk space by removing FRC C++/Java programs. Fails if not ran as admin.
    """

    logger.info("Clearing FRC C++/Java program support")

    rm_paths = (third_party_libs, static_deploy)

    ssh.exec_bash(
        # Remove java ipk
        'opkg remove "frc*-openjdk*"',
        # Remove third party libs not used by RobotPy
        f"rm -rf {' '.join(rm_paths)}",
        bash_opts="ex",
        print_output=True,
        check=True,
    )


def get_rio_py_packages(ssh: SshController) -> typing.Dict[str, str]:
    # Use importlib.metadata instead of pip because it's way faster than pip
    result = ssh.exec_cmd(
        "/usr/local/bin/python3 -c "
        "'from importlib.metadata import distributions;"
        "import json; import sys; "
        "json.dump({dist.name: dist.version for dist in distributions()},sys.stdout)'",
        get_output=True,
    )
    assert result.stdout is not None
    return json.loads(result.stdout)


def get_kill_script() -> bytes:
    global kill_script_content
    if kill_script_content is None:
        with open(pathlib.Path(__file__).parent / "frcKillRobot.sh", "rb") as fp:
            kill_script_content = fp.read()

    return kill_script_content


def check_kill_script(ssh: SshController) -> bool:
    ks: bytes = get_kill_script()
    ks_hash = hashlib.md5(ks).hexdigest()
    output = ssh.check_output(f"md5sum {kill_robot_script}")
    return output.split(" ", 1)[0] == ks_hash


def update_kill_script(ssh: SshController):
    logger.info("Updating %s", kill_robot_script)
    fp = io.BytesIO(get_kill_script())
    ssh.sftp_fp(fp, kill_robot_script)



import contextlib
import inspect
import io
import json
import logging
import pathlib
import re
import shlex
import subprocess
import sys
from urllib.parse import urlparse
import typing

from os.path import basename, exists

from packaging.version import Version

from .version import version as __version__
from . import roborio_utils
from .cacheserver import CacheServer
from .errors import Error, SshExecError
from .sshcontroller import SshController, ssh_from_cfg
from .utils import _urlretrieve

_WPILIB_YEAR = "2025"
_IS_BETA = False

_ROBORIO_WHEELS = f"https://wpilib.jfrog.io/artifactory/api/pypi/wpilib-python-release-{_WPILIB_YEAR}/simple"

_ROBORIO_IMAGES = [
    "2025_v2.0",
]

_ROBORIO2_IMAGES = [
    "2025_v2.0",
]

_ROBOTPY_PYTHON_PLATFORM = "linux_roborio"
_ROBOTPY_PYTHON_VERSION_NUM = "313"
_ROBOTPY_PYTHON_VERSION = f"python{_ROBOTPY_PYTHON_VERSION_NUM}"

_PIP_STUB_PATH = "/home/admin/rpip"

_PYTHON_IPK = "https://github.com/robotpy/roborio-python/releases/download/2025-3.13.1-r1/python313_3.13.1-r1_cortexa9-vfpv3.ipk"

logger = logging.getLogger("robotpy.installer")


class InstallerException(Error):
    pass


class PipInstallError(InstallerException):
    pass


class PythonMissingError(InstallerException):
    pass


@contextlib.contextmanager
def catch_ssh_error(msg: str):
    try:
        yield
    except SshExecError as e:
        raise InstallerException(f"{msg}: {e}")


class RobotpyInstaller:
    def __init__(self, *, log_startup: bool = True):
        self.cache_root = pathlib.Path.home() / "wpilib" / _WPILIB_YEAR / "robotpy"
        self.pip_cache = self.cache_root / "pip_cache"
        self.opkg_cache = self.cache_root / "opkg_cache"

        self._ssh: typing.Optional[SshController] = None
        self._cache_server: typing.Optional[CacheServer] = None

        self._image_version_ok = False
        self._robot_pip_ok = False

        self._webserver_stopped = False
        self._webserver_needs_start = False

        if log_startup:
            logger.info("RobotPy Installer %s", __version__)
            logger.info("-> caching files at %s", self.cache_root)

    @contextlib.contextmanager
    def connect_to_robot(
        self,
        *,
        project_path: pathlib.Path,
        main_file: pathlib.Path,
        robot_or_team: typing.Union[None, str, int] = None,
        ignore_image_version: bool = False,
        log_usage: bool = True,
        no_resolve: bool = False,
        ssh: typing.Optional[SshController] = None,
    ):
        if ssh is None:
            ssh = ssh_from_cfg(
                project_path,
                main_file,
                username="admin",
                password="",
                robot_or_team=robot_or_team,
                no_resolve=no_resolve,
            )
        elif ssh.username != "admin":
            ssh = SshController(ssh.hostname, "admin", "")

        with ssh:
            self._ssh = ssh

            self.ensure_image_version(ignore_image_version)

            if log_usage:
                self.show_disk_space()
                self.show_mem_usage()

            yield

            if self._webserver_needs_start:
                self.ssh.exec_cmd("/etc/init.d/systemWebServer start")
                self._webserver_needs_start = False
                self._webserver_stopped = False

            if log_usage:
                self.show_disk_space()
                self.show_mem_usage()

            self._ssh = None

    @property
    def cache_server(self) -> CacheServer:
        """Only access inside connect_to_robot context"""
        if not self._cache_server:
            self._cache_server = CacheServer(self.ssh, self.cache_root)
            self._cache_server.start()

        return self._cache_server

    @property
    def ssh(self) -> SshController:
        """Only access inside connect_to_robot context"""
        if self._ssh is None:
            raise RuntimeError("internal error")
        return self._ssh

    #
    # Utilities
    #

    def opkg_install(
        self,
        force_reinstall: bool,
        packages: typing.Sequence[pathlib.Path],
    ):
        """
        Installs opkg package on RoboRIO
        """

        for package in packages:
            if package.parent != self.opkg_cache:
                raise ValueError("internal error")
            if not package.exists():
                raise PythonMissingError(
                    f"{package.name} has not been downloaded yet\n"
                    "- Use 'python -m robotpy installer download-python' to download"
                )

        # Write out the install script
        # -> we use a script because opkg doesn't have a good mechanism
        #    to only install a package if it's not already installed
        opkg_files = []

        opkg_script = inspect.cleandoc(
            """
            set -e
            PACKAGES=()
            DO_INSTALL=0
            """
        )

        opkg_script_bit = inspect.cleandoc(
            f"""
            if ! opkg list-installed | grep -F "%(name)s - %(version)s"; then
                PACKAGES+=("http://localhost:{self.cache_server.port}/opkg_cache/%(fname)s")
                DO_INSTALL=1
            else
                echo "%(name)s already installed"
            fi
            """
        )

        for package in packages:
            pkgname, pkgversion, _ = package.name.split("_")

            opkg_script += "\n" + (
                opkg_script_bit
                % {
                    "fname": package.name,
                    "name": pkgname,
                    "version": pkgversion,
                }
            )

            opkg_files.append(package.name)

        # Finish it out
        opkg_script += "\n" + (
            inspect.cleandoc(
                """
                if [ "${DO_INSTALL}" == "0" ]; then
                    echo "No packages to install."
                else
                    echo + opkg install %(options)s ${PACKAGES[@]}
                    opkg install %(options)s ${PACKAGES[@]}
                fi

                sync
                ldconfig
                """
            )
            % {"options": "--force-reinstall" if force_reinstall else ""}
        )

        with catch_ssh_error("creating opkg install script"):
            # write to /tmp so that it doesn't persist
            self.ssh.exec_cmd(
                f"echo '{opkg_script}' > /tmp/install_opkg.sh",
                check=True,
            )

        with catch_ssh_error("installing selected packages"):
            self.ssh.exec_cmd(
                "bash /tmp/install_opkg.sh", check=True, print_output=True
            )

        try:
            self.ssh.exec_cmd("rm /tmp/install_opkg.sh")
        except SshExecError:
            pass

    def show_disk_space(
        self,
    ) -> typing.Tuple[str, str, str]:
        #
        # Free space check.. maybe in the future we'll use this to not accidentally
        # fill the user's disk, but it'd be annoying to figure out
        #

        with catch_ssh_error("checking free space"):
            result = self.ssh.check_output("df -h / | tail -n 1")

        _, size, used, _, pct, _ = result.strip().split()
        logger.info("-> RoboRIO disk usage %s/%s (%s full)", used, size, pct)

        return size, used, pct

    def show_mem_usage(self):
        with catch_ssh_error("checking memory info"):
            result = self.ssh.check_output("cat /proc/meminfo")

        total_kb = 0
        available_kb = 0
        found = 0

        for line in result.strip().splitlines():
            if line.startswith("MemTotal:"):
                total_kb = int(line.split()[1])
                found += 1
            elif line.startswith("MemAvailable"):
                available_kb = int(line.split()[1])
                found += 1

            if found == 2:
                break

        used_kb = total_kb - available_kb
        pct_free = (available_kb / float(total_kb)) * 100.0

        logger.info(
            "-> RoboRIO memory %.1fM/%.1fM (%.0f%% full)",
            used_kb / 1000.0,
            total_kb / 1000.0,
            pct_free,
        )

    def ensure_more_memory(self):
        if self._webserver_stopped:
            return

        # This takes up a ton of memory and we need the memory...
        with catch_ssh_error("Stopping NI webserver"):
            result = self.ssh.exec_bash('[ -z "$(ps | grep NIWebServiceContainer)" ]')
            if result.returncode != 0:
                self.ssh.exec_cmd("/etc/init.d/systemWebServer stop")
                self._webserver_needs_start = True

        self._webserver_stopped = True

    def ensure_image_version(self, ignore_image_version: bool):
        if self._image_version_ok:
            return

        with catch_ssh_error("retrieving image version"):
            result = self.ssh.check_output(
                "grep IMAGEVERSION /etc/natinst/share/scs_imagemetadata.ini",
            )

        roborio_match = re.match(
            r'IMAGEVERSION = "(FRC_)?roboRIO_(.*)"', result.strip()
        )
        roborio2_match = re.match(
            r'IMAGEVERSION = "(FRC_)?roboRIO2_(.*)"', result.strip()
        )

        if roborio_match:
            version = roborio_match.group(2)
            images = _ROBORIO_IMAGES
            name = "RoboRIO"
        elif roborio2_match:
            version = roborio2_match.group(2)
            images = _ROBORIO2_IMAGES
            name = "RoboRIO 2"
        else:
            version = "<unknown>"
            images = [
                f"({_ROBORIO_IMAGES[-1]} | {_ROBORIO2_IMAGES[-1]})",
            ]
            name = "RoboRIO (1 | 2)"

        logger.info(f"-> {name} image version: {version}")

        if not ignore_image_version and version not in images:
            raise InstallerException(
                f"{name} image {images[-1]} is required!\n"
                "\n"
                "See https://docs.wpilib.org/en/stable/docs/zero-to-robot/step-3/imaging-your-roborio.html\n"
                "for information about upgrading the RoboRIO image.\n"
                "\n"
                "Use --ignore-image-version to install anyways"
            )

        self._image_version_ok = True

    def ensure_robot_pip(self):
        if self._robot_pip_ok:
            return

        #
        # Ensure that pip is installed
        #

        with catch_ssh_error("checking for pip3"):
            if self.ssh.exec_cmd("[ -x /usr/local/bin/pip3 ]").returncode != 0:
                raise InstallerException(
                    inspect.cleandoc(
                        """
                        pip3 not found on RoboRIO, did you install python?

                        Use the 'download-python' and 'install-python' commands first!
                        """
                    )
                )

        # Use pip stub to override the wheel platform on roborio
        with catch_ssh_error("copying pip stub"):
            from . import _pipstub

            stub_fp = io.BytesIO()
            stub_fp.write(b"#!/usr/local/bin/python3\n\n")
            stub_fp.write(inspect.getsource(_pipstub).encode("utf-8"))
            stub_fp.seek(0)

            self.ssh.sftp_fp(stub_fp, _PIP_STUB_PATH)
            self.ssh.exec_cmd(f"chmod +x {_PIP_STUB_PATH}", check=True)

        self._robot_pip_ok = True

    #
    # Python installation
    #

    @property
    def _python_ipk_path(self) -> pathlib.Path:
        parts = urlparse(_PYTHON_IPK)
        return self.opkg_cache / pathlib.PurePosixPath(parts.path).name

    def is_python_downloaded(self) -> bool:
        return self._python_ipk_path.exists()

    def download_python(self, use_certifi: bool):
        self.opkg_cache.mkdir(parents=True, exist_ok=True)

        ipk_dst = self._python_ipk_path
        _urlretrieve(_PYTHON_IPK, ipk_dst, True, _make_ssl_context(use_certifi))

    def install_python(self):
        """
        Installs Python on a RoboRIO.

        Requires download-python to be executed first.
        """
        logger.info("Installing Python on RoboRIO (this may take a few minutes)")
        ipk_dst = self._python_ipk_path
        self.opkg_install(False, [ipk_dst])

    def uninstall_python(
        self,
    ):
        with catch_ssh_error("removing python"):
            self.ssh.exec_cmd(
                f"opkg remove {_ROBOTPY_PYTHON_VERSION}",
                check=True,
                print_output=True,
            )

    def uninstall_robotpy(self):
        with catch_ssh_error("removing user program"):
            self.ssh.exec_bash(
                roborio_utils.kill_robot_cmd,
                "rm -rf /home/lvuser/py",
                f"rm -f {roborio_utils.robot_command}",
            )

        with catch_ssh_error("removing pip packages"):
            pkgs = roborio_utils.get_rio_py_packages(self.ssh)
            self.pip_uninstall([pkg for pkg in pkgs.keys() if pkg != "pip"])

        self.uninstall_python()

    #
    # pip packages
    #

    def _extend_pip_args(
        self,
        pip_args: typing.List[str],
        cache: typing.Optional[CacheServer],
        force_reinstall: bool,
        ignore_installed: bool,
        no_deps: bool,
        pre: bool,
        requirements: typing.Iterable[pathlib.Path],
    ):
        if pre:
            pip_args.append("--pre")
        if force_reinstall:
            pip_args.append("--force-reinstall")
        if ignore_installed:
            pip_args.append("--ignore-installed")
        if no_deps:
            pip_args.append("--no-deps")

        for req in requirements:
            if cache:
                fname = f"/requirements/{req.name}"
                cache.add_mapping(fname, str(req))
                pip_args.extend(["-r", f"http://localhost:{cache.port}{fname}"])
            else:
                pip_args.extend(["-r", str(req)])

    def pip_download(
        self,
        no_deps: bool,
        pre: bool,
        requirements: typing.Sequence[pathlib.Path],
        packages: typing.Sequence[str],
    ):
        """
        Specify Python package(s) to download, and store them in the cache.

        You must be connected to the internet for this to work.
        """

        if not requirements and not packages:
            raise InstallerException(
                "You must give at least one requirement to download"
            )

        try:
            import pip  # type: ignore
        except ImportError:
            raise InstallerException(
                "ERROR: pip must be installed to download python packages"
            )

        self.pip_cache.mkdir(parents=True, exist_ok=True)

        pip_args = [
            "--no-cache-dir",
            "--disable-pip-version-check",
            "download",
            "--extra-index-url",
            _ROBORIO_WHEELS,
            "--only-binary",
            ":all:",
            "--platform",
            _ROBOTPY_PYTHON_PLATFORM,
            "--python-version",
            _ROBOTPY_PYTHON_VERSION_NUM,
            "--implementation",
            "cp",
            "--abi",
            f"cp{_ROBOTPY_PYTHON_VERSION_NUM}",
            "-d",
            str(self.pip_cache),
        ]

        self._extend_pip_args(
            pip_args,
            None,
            False,
            False,
            no_deps,
            pre,
            requirements,
        )

        pip_args.extend(packages)
        pip_args = [sys.executable, "-m", "robotpy_installer._pipstub"] + pip_args

        logger.debug("Using pip to download: %s", pip_args)

        retval = subprocess.call(pip_args)
        if retval != 0:
            raise InstallerException("pip download failed")

    def pip_install(
        self,
        force_reinstall: bool,
        ignore_installed: bool,
        no_deps: bool,
        pre: bool,
        requirements: typing.Sequence[pathlib.Path],
        packages: typing.Sequence[str],
    ):
        """
        Installs Python package(s) on a RoboRIO.

        The package must already been downloaded with the 'download' command first.
        """

        self.ensure_robot_pip()

        if len(requirements) == 0 and len(packages) == 0:
            raise InstallerException(
                "You must give at least one requirement to install"
            )

        cache_server = self.cache_server

        pip_args = [
            "/home/admin/rpip",
            "--no-cache-dir",
            "--disable-pip-version-check",
            "install",
            "--no-index",
            "--root-user-action=ignore",
            "--find-links",
            f"http://localhost:{cache_server.port}/pip_cache/",
            # always add --upgrade, anything in the cache should be installed
            "--upgrade",
            "--upgrade-strategy=eager",
        ]

        self._extend_pip_args(
            pip_args,
            cache_server,
            force_reinstall,
            ignore_installed,
            no_deps,
            pre,
            requirements,
        )

        for package in packages:
            if package.endswith(".whl") and exists(package):
                fname = basename(package)
                cache_server.add_mapping(f"/extra/{fname}", package)
                pip_args.append(f"http://localhost:{cache_server.port}/extra/{fname}")
            else:
                pip_args.append(package)

        # pip is greedy
        self.ensure_more_memory()

        try:
            self.ssh.exec_cmd(shlex.join(pip_args), check=True, print_output=True)
        except SshExecError as e:
            raise PipInstallError(f"installing packages: {e}") from e

        # Some of our hacky wheels require this
        with catch_ssh_error("running ldconfig"):
            self.ssh.exec_cmd("ldconfig")

    def pip_list(self):
        self.ensure_robot_pip()

        # pip is greedy
        self.ensure_more_memory()

        with catch_ssh_error("pip3 list"):
            self.ssh.exec_cmd(
                f"{_PIP_STUB_PATH} --no-cache-dir --disable-pip-version-check list",
                check=True,
                print_output=True,
            )

    def pip_uninstall(
        self,
        packages: typing.Sequence[str],
    ):
        self.ensure_robot_pip()

        if len(packages) == 0:
            raise InstallerException("You must give at least one package to uninstall")

        pip_args = [
            _PIP_STUB_PATH,
            "--no-cache-dir",
            "--disable-pip-version-check",
            "uninstall",
            "--root-user-action=ignore",
            "--yes",
        ]
        pip_args.extend(packages)

        # pip is greedy
        self.ensure_more_memory()

        with catch_ssh_error("uninstalling packages"):
            self.ssh.exec_cmd(shlex.join(pip_args), check=True, print_output=True)

    def get_pypi_version(self, package: str, use_certifi: bool) -> Version:
        """
        Retrieves the latest version of a package on pypi that corresponds to the current year
        """
        self.cache_root.mkdir(parents=True, exist_ok=True)
        fname = self.cache_root / f"pypi-{package}.json"
        _urlretrieve(
            f"https://pypi.org/simple/{package}",
            fname,
            True,
            _make_ssl_context(use_certifi),
            False,
            {"Accept": "application/vnd.pypi.simple.v1+json"},
        )
        with open(fname, "r") as fp:
            data = json.load(fp)

        versions = [Version(v) for v in data["versions"]]

        # Sort the versions
        maxv = Version(str(int(_WPILIB_YEAR) + 1))

        def _version_ok(v: Version) -> bool:
            ok = v < maxv and not v.is_devrelease
            if ok and not _IS_BETA:
                ok = not v.is_prerelease
            return ok

        versions = sorted(v for v in versions if _version_ok(v))
        if not versions:
            raise InstallerException(f"could not find {package} version on pypi")

        return versions[-1]


def _make_ssl_context(use_certifi: bool):
    if not use_certifi:
        return None

    try:
        import certifi  # type: ignore
    except ImportError:
        raise InstallerException(
            "certifi is not installed, please install it via `pip install certifi`"
        )

    import ssl

    return ssl.create_default_context(ssl.Purpose.SERVER_AUTH, cafile=certifi.where())


def main():
    print("ERROR: robotpy-installer is now a subcommand of 'robotpy'", file=sys.stderr)
    print("- Use 'python -m robotpy installer'", file=sys.stderr)
    sys.exit(1)
